"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.any = void 0;
exports.anyOf = anyOf;
exports.characterClass = characterClass;
exports.characterRange = characterRange;
exports.digit = void 0;
exports.inverted = inverted;
exports.word = exports.whitespace = void 0;
/**
 * Character range from start to end (inclusive).
 */

const any = exports.any = {
  type: 'characterClass',
  characters: ['.'],
  ranges: [],
  isInverted: false,
  encode: encodeCharacterClass
};
const digit = exports.digit = {
  type: 'characterClass',
  characters: ['\\d'],
  ranges: [],
  isInverted: false,
  encode: encodeCharacterClass
};
const word = exports.word = {
  type: 'characterClass',
  characters: ['\\w'],
  ranges: [],
  isInverted: false,
  encode: encodeCharacterClass
};
const whitespace = exports.whitespace = {
  type: 'characterClass',
  characters: ['\\s'],
  ranges: [],
  isInverted: false,
  encode: encodeCharacterClass
};
function characterClass(...elements) {
  elements.forEach(element => {
    if (element.isInverted) {
      throw new Error('`characterClass` should receive only non-inverted character classes');
    }
  });
  return {
    type: 'characterClass',
    characters: elements.map(c => c.characters).flat(),
    ranges: elements.map(c => c.ranges).flat(),
    isInverted: false,
    encode: encodeCharacterClass
  };
}
function characterRange(start, end) {
  if (start.length !== 1) {
    throw new Error('`characterRange` should receive only single character `start` string');
  }
  if (end.length !== 1) {
    throw new Error('`characterRange` should receive only single character `end` string');
  }
  if (start > end) {
    throw new Error('`start` should be before or equal to `end`');
  }
  return {
    type: 'characterClass',
    characters: [],
    ranges: [{
      start,
      end
    }],
    isInverted: false,
    encode: encodeCharacterClass
  };
}
function anyOf(characters) {
  const charactersArray = characters.split('').map(c => escapeForCharacterClass(c));
  if (charactersArray.length === 0) {
    throw new Error('`anyOf` should received at least one character');
  }
  return {
    type: 'characterClass',
    characters: charactersArray,
    ranges: [],
    isInverted: false,
    encode: encodeCharacterClass
  };
}
function inverted(element) {
  return {
    type: 'characterClass',
    characters: element.characters,
    ranges: element.ranges,
    isInverted: !element.isInverted,
    encode: encodeCharacterClass
  };
}
function encodeCharacterClass() {
  var _this$ranges;
  if (this.characters.length === 0 && this.ranges.length === 0) {
    throw new Error('Character class should contain at least one character or character range');
  }

  // Direct rendering for single-character class
  if (this.characters.length === 1 && ((_this$ranges = this.ranges) === null || _this$ranges === void 0 ? void 0 : _this$ranges.length) === 0 && !this.isInverted) {
    return {
      precedence: 'atom',
      pattern: this.characters[0]
    };
  }

  // If passed characters includes hyphen (`-`) it need to be moved to
  // first (or last) place in order to treat it as hyphen character and not a range.
  // See: https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_expressions/Character_classes#types
  const hyphen = this.characters.includes('-') ? '-' : '';
  const otherCharacters = this.characters.filter(c => c !== '-').join('');
  const ranges = this.ranges.map(({
    start,
    end
  }) => `${start}-${end}`).join('');
  const isInverted = this.isInverted ? '^' : '';
  return {
    precedence: 'atom',
    pattern: `[${isInverted}${ranges}${otherCharacters}${hyphen}]`
  };
}
function escapeForCharacterClass(text) {
  return text.replace(/[\]\\]/g, '\\$&'); // $& means the whole matched string
}
//# sourceMappingURL=character-class.js.map